/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "quakedef.h"
#include "particles.h"


emitter_t *free_emitters = NULL;
emitter_t *active_emitters = NULL;
particle_t *free_particles = NULL;

// ramps are made larger to allow for overshoot, using full alpha as the dead particle
int	ramp1[] = {0x6f, 0x6d, 0x6b, 0x69, 0x67, 0x65, 0x63, 0x61, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
int	ramp2[] = {0x6f, 0x6e, 0x6d, 0x6c, 0x6b, 0x6a, 0x68, 0x66, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
int	ramp3[] = {0x6d, 0x6b, 0x06, 0x05, 0x04, 0x03, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff};


emitter_t *R_AllocEmitter (float *origin)
{
	emitter_t *e;

	if (!free_emitters)
	{
		// if we run out of emitters just alloc another (this will go in the per-map hunk and will be reset on load)
		free_emitters = (emitter_t *) Hunk_Alloc (sizeof (emitter_t));
		free_emitters->next = NULL;
	}

	// take an emitter from the free list
	e = free_emitters;

	free_emitters = e->next;
	e->next = active_emitters;
	active_emitters = e;

	// set the time that the emitter was spawned at for velocity-based movement
	// (all particles in the emitter will be spawned at the same time)
	e->time = cl.time;

	// fill in other properties
	e->particles = NULL;
	Vector3Copy (e->origin, origin);

	return e;
}


particle_t *R_AllocParticle (emitter_t *e)
{
	particle_t *p;

	if (!free_particles)
	{
		free_particles = (particle_t *) Hunk_Alloc (sizeof (particle_t));
		free_particles->next = NULL;
	}

	p = free_particles;

	free_particles = p->next;
	p->next = e->particles;
	e->particles = p;

	return p;
}


/*
===============
R_ClearParticles
===============
*/
void R_ClearParticles (void)
{
	free_emitters = NULL;
	active_emitters = NULL;
	free_particles = NULL;
}


/*
===============
R_EntityParticles
===============
*/

#define NUMVERTEXNORMALS	162
extern	float	r_avertexnormals[NUMVERTEXNORMALS][3];
vec3_t	avelocities[NUMVERTEXNORMALS];
float	beamlength = 16;
vec3_t	avelocity = {23, 7, 3};
float	partstep = 0.01;
float	timescale = 0.01;

void R_EntityParticles (entity_t *ent)
{
	int			count;
	int			i;
	particle_t	*p;
	float		angle;
	float		sr, sp, sy, cr, cp, cy;
	vec3_t		forward;
	float		dist;
	emitter_t	*e = R_AllocEmitter (ent->origin);

	dist = 64;
	count = 50;

	if (!avelocities[0][0])
	{
		for (i = 0; i < NUMVERTEXNORMALS * 3; i++)
			avelocities[0][i] = (rand () & 255) * 0.01;
	}

	for (i = 0; i < NUMVERTEXNORMALS; i++)
	{
		angle = cl.time * avelocities[i][0];
		sy = sin (angle);
		cy = cos (angle);
		angle = cl.time * avelocities[i][1];
		sp = sin (angle);
		cp = cos (angle);
		angle = cl.time * avelocities[i][2];
		sr = sin (angle);
		cr = cos (angle);

		forward[0] = cp*cy;
		forward[1] = cp*sy;
		forward[2] = -sp;

		if ((p = R_AllocParticle (e)) == NULL)
			return;

		p->die = cl.time + 0.01;
		p->color = 0x6f;
		p->type = pt_explode;

		p->org[0] = ent->origin[0] + r_avertexnormals[i][0] * dist + forward[0] * beamlength;
		p->org[1] = ent->origin[1] + r_avertexnormals[i][1] * dist + forward[1] * beamlength;
		p->org[2] = ent->origin[2] + r_avertexnormals[i][2] * dist + forward[2] * beamlength;
	}
}


void R_ReadPointFile_f (void)
{
	FILE	*f;
	vec3_t	org;
	int		r;
	int		c;
	particle_t	*p;
	char	name[MAX_OSPATH];
	emitter_t	*e;

	sprintf (name, "maps/%s.pts", sv.name);
	COM_FOpenFile (name, &f);

	if (!f)
	{
		Con_Printf ("couldn't open %s\n", name);
		return;
	}

	Con_Printf ("Reading %s...\n", name);
	c = 0;

	e = R_AllocEmitter (vec3_origin);

	for (;;)
	{
		r = fscanf (f, "%f %f %f\n", &org[0], &org[1], &org[2]);

		if (r != 3)
			break;
		c++;

		if ((p = R_AllocParticle (e)) == NULL)
		{
			Con_Printf ("Not enough free particles\n");
			break;
		}

		p->die = 99999;
		p->color = (-c) & 15;
		p->type = pt_static;

		VectorCopy (vec3_origin, p->vel);
		VectorCopy (org, p->org);
	}

	fclose (f);
	Con_Printf ("%i points read\n", c);
}


/*
===============
R_ParseParticleEffect

Parse an effect out of the server message
===============
*/
void R_ParseParticleEffect (void)
{
	vec3_t		org, dir;
	int			i, count, msgcount, color;

	for (i = 0; i < 3; i++)
		org[i] = MSG_ReadCoord ();
	for (i = 0; i < 3; i++)
		dir[i] = MSG_ReadChar () * (1.0 / 16);
	msgcount = MSG_ReadByte ();
	color = MSG_ReadByte ();

	if (msgcount == 255)
		count = 1024;
	else
		count = msgcount;

	R_RunParticleEffect (org, dir, color, count);
}

/*
===============
R_ParticleExplosion

===============
*/
void R_ParticleExplosion (vec3_t org)
{
	int			i, j;
	particle_t	*p;
	emitter_t	*e = R_AllocEmitter (org);

	for (i = 0; i < 1024; i++)
	{
		if ((p = R_AllocParticle (e)) == NULL)
			return;

		p->die = cl.time + 5;
		p->color = ramp1[0];
		p->ramp = rand () & 3;
		if (i & 1)
		{
			p->type = pt_explode;
			for (j = 0; j < 3; j++)
			{
				p->org[j] = org[j] + ((rand () % 32) - 16);
				p->vel[j] = (rand () % 512) - 256;
			}
		}
		else
		{
			p->type = pt_explode2;
			for (j = 0; j < 3; j++)
			{
				p->org[j] = org[j] + ((rand () % 32) - 16);
				p->vel[j] = (rand () % 512) - 256;
			}
		}
	}
}

/*
===============
R_ParticleExplosion2

===============
*/
void R_ParticleExplosion2 (vec3_t org, int colorStart, int colorLength)
{
	int			i, j;
	particle_t	*p;
	int			colorMod = 0;
	emitter_t	*e = R_AllocEmitter (org);

	for (i = 0; i < 512; i++)
	{
		if ((p = R_AllocParticle (e)) == NULL)
			return;

		p->die = cl.time + 0.3;
		p->color = colorStart + (colorMod % colorLength);
		colorMod++;

		p->type = pt_blob;
		for (j = 0; j < 3; j++)
		{
			p->org[j] = org[j] + ((rand () % 32) - 16);
			p->vel[j] = (rand () % 512) - 256;
		}
	}
}

/*
===============
R_BlobExplosion

===============
*/
void R_BlobExplosion (vec3_t org)
{
	int			i, j;
	particle_t	*p;
	emitter_t	*e = R_AllocEmitter (org);

	for (i = 0; i < 1024; i++)
	{
		if ((p = R_AllocParticle (e)) == NULL)
			return;

		p->die = cl.time + 1 + (rand () & 8) * 0.05;

		if (i & 1)
		{
			p->type = pt_blob;
			p->color = 66 + rand () % 6;
			for (j = 0; j < 3; j++)
			{
				p->org[j] = org[j] + ((rand () % 32) - 16);
				p->vel[j] = (rand () % 512) - 256;
			}
		}
		else
		{
			p->type = pt_blob2;
			p->color = 150 + rand () % 6;
			for (j = 0; j < 3; j++)
			{
				p->org[j] = org[j] + ((rand () % 32) - 16);
				p->vel[j] = (rand () % 512) - 256;
			}
		}
	}
}

/*
===============
R_RunParticleEffect

===============
*/
void R_RunParticleEffect (vec3_t org, vec3_t dir, int color, int count)
{
	if (count == 1024)
	{
		R_ParticleExplosion (org);
		return;
	}
	else
	{
		int			i, j;
		particle_t	*p;
		emitter_t	*e = R_AllocEmitter (org);

		for (i = 0; i < count; i++)
		{
			if ((p = R_AllocParticle (e)) == NULL)
				return;

			p->die = cl.time + 0.1 * (rand () % 5);
			p->color = (color & ~7) + (rand () & 7);
			p->type = pt_slowgrav;

			for (j = 0; j < 3; j++)
			{
				p->org[j] = org[j] + ((rand () & 15) - 8);
				p->vel[j] = dir[j] * 15;
			}
		}
	}
}


/*
===============
R_LavaSplash

===============
*/
void R_LavaSplash (vec3_t org)
{
	int			i, j, k;
	particle_t	*p;
	float		vel;
	vec3_t		dir;
	emitter_t	*e = R_AllocEmitter (org);

	for (i = -16; i < 16; i++)
	{
		for (j = -16; j < 16; j++)
		{
			for (k = 0; k < 1; k++)
			{
				if ((p = R_AllocParticle (e)) == NULL)
					return;

				p->die = cl.time + 2 + (rand () & 31) * 0.02;
				p->color = 224 + (rand () & 7);
				p->type = pt_slowgrav;

				dir[0] = j * 8 + (rand () & 7);
				dir[1] = i * 8 + (rand () & 7);
				dir[2] = 256;

				p->org[0] = org[0] + dir[0];
				p->org[1] = org[1] + dir[1];
				p->org[2] = org[2] + (rand () & 63);

				VectorNormalize (dir);
				vel = 50 + (rand () & 63);
				VectorScale (dir, vel, p->vel);
			}
		}
	}
}

/*
===============
R_TeleportSplash

===============
*/
void R_TeleportSplash (vec3_t org)
{
	int			i, j, k;
	particle_t	*p;
	float		vel;
	vec3_t		dir;
	emitter_t	*e = R_AllocEmitter (org);

	for (i = -16; i < 16; i += 4)
	{
		for (j = -16; j < 16; j += 4)
		{
			for (k = -24; k < 32; k += 4)
			{
				if ((p = R_AllocParticle (e)) == NULL)
					return;

				p->die = cl.time + 0.2 + (rand () & 7) * 0.02;
				p->color = 7 + (rand () & 7);
				p->type = pt_slowgrav;

				dir[0] = j * 8;
				dir[1] = i * 8;
				dir[2] = k * 8;

				p->org[0] = org[0] + i + (rand () & 3);
				p->org[1] = org[1] + j + (rand () & 3);
				p->org[2] = org[2] + k + (rand () & 3);

				VectorNormalize (dir);
				vel = 50 + (rand () & 63);
				VectorScale (dir, vel, p->vel);
			}
		}
	}
}

void R_RocketTrail (vec3_t start, vec3_t end, int type)
{
	vec3_t		vec;
	float		len;
	int			j;
	particle_t	*p;
	int			dec;
	static int	tracercount;
	emitter_t	*e;

	VectorSubtract (end, start, vec);
	len = VectorNormalize (vec);

	if (type < 128)
		dec = 3;
	else
	{
		dec = 1;
		type -= 128;
	}

	e = R_AllocEmitter (start);

	while (len > 0)
	{
		len -= dec;

		if ((p = R_AllocParticle (e)) == NULL)
			return;

		VectorCopy (vec3_origin, p->vel);
		p->die = cl.time + 2;

		switch (type)
		{
		case 0:	// rocket trail
			p->ramp = (rand () & 3);
			p->color = ramp3[(int) p->ramp];
			p->type = pt_fire;
			for (j = 0; j < 3; j++)
				p->org[j] = start[j] + ((rand () % 6) - 3);
			break;

		case 1:	// smoke smoke
			p->ramp = (rand () & 3) + 2;
			p->color = ramp3[(int) p->ramp];
			p->type = pt_fire;
			for (j = 0; j < 3; j++)
				p->org[j] = start[j] + ((rand () % 6) - 3);
			break;

		case 2:	// blood
			p->type = pt_grav;
			p->color = 67 + (rand () & 3);
			for (j = 0; j < 3; j++)
				p->org[j] = start[j] + ((rand () % 6) - 3);
			break;

		case 3:
		case 5:	// tracer
			p->die = cl.time + 0.5;
			p->type = pt_static;
			if (type == 3)
				p->color = 52 + ((tracercount & 4) << 1);
			else
				p->color = 230 + ((tracercount & 4) << 1);

			tracercount++;

			VectorCopy (start, p->org);

			if (tracercount & 1)
			{
				p->vel[0] = 30 * vec[1];
				p->vel[1] = 30 * -vec[0];
			}
			else
			{
				p->vel[0] = 30 * -vec[1];
				p->vel[1] = 30 * vec[0];
			}
			break;

		case 4:	// slight blood
			p->type = pt_grav;
			p->color = 67 + (rand () & 3);
			for (j = 0; j < 3; j++)
				p->org[j] = start[j] + ((rand () % 6) - 3);
			len -= 3;
			break;

		case 6:	// voor trail
			p->color = 9 * 16 + 8 + (rand () & 3);
			p->type = pt_static;
			p->die = cl.time + 0.3;
			for (j = 0; j < 3; j++)
				p->org[j] = start[j] + ((rand () & 15) - 8);
			break;
		}

		VectorAdd (start, vec, start);
	}
}


void CL_RunEmitter (emitter_t *e)
{
	float etime = cl.time - e->time;
	particle_t *p;
	ptypedef_t *pt;
	int i;

	// ??? should never happen ???
	if (!e->particles) return;

	e->mins[0] = e->mins[1] = e->mins[2] = 9999999.0f;
	e->maxs[0] = e->maxs[1] = e->maxs[2] = -9999999.0f;

	// move dead particles at the start of the list back to free
	for (;;)
	{
		particle_t *kill = e->particles;

		if (kill && kill->die < cl.time)
		{
			e->particles = kill->next;
			kill->next = free_particles;
			free_particles = kill;
			continue;
		}

		break;
	}

	for (p = e->particles; p; p = p->next)
	{
		// move dead particles in the middle of the list back to free
		for (;;)
		{
			particle_t *kill = p->next;

			if (kill && kill->die < cl.time)
			{
				p->next = kill->next;
				kill->next = free_particles;
				free_particles = kill;
				continue;
			}

			break;
		}

		// get the emitter properties for this particle
		pt = &p_typedefs[p->type];

		// move the particle in a framerate-independent manner
		Vector3Set (
			p->move,
			p->org[0] + (p->vel[0] + (pt->accel[0] * etime)) * etime,
			p->org[1] + (p->vel[1] + (pt->accel[1] * etime)) * etime,
			p->org[2] + (p->vel[2] + (pt->accel[2] * etime)) * etime
		);

		// accumulate emitter bbox
		for (i = 0; i < 3; i++)
		{
			if (p->move[i] < e->mins[i]) e->mins[i] = p->move[i];
			if (p->move[i] > e->maxs[i]) e->maxs[i] = p->move[i];
		}

		// update colour ramps
		if (pt->ramp)
		{
			int ramp = (int) (p->ramp + (etime * pt->ramptime));

			// set dead particles to full-alpha and the system will remove them on the next frame
			if (ramp > 8)
			{
				p->die = -1;
				continue;
			}
			else if ((p->color = pt->ramp[ramp]) == 0xff)
			{
				p->die = -1;
				continue;
			}
		}
	}
}


void CL_RunParticles (void)
{
	int i;
	emitter_t *e;

	// update particle accelerations
	for (i = 0; i < MAX_PARTICLE_TYPES; i++)
	{
		extern cvar_t sv_gravity;
		ptypedef_t *pt = &p_typedefs[i];
		float grav = sv_gravity.value * 0.05;

		// in theory this could be calced once and never again, but in practice mods may change sv_gravity from frame-to-frame
		// so we need to recalc it each frame too....
		pt->accel[0] = pt->dvel[0] + (pt->grav[0] * grav);
		pt->accel[1] = pt->dvel[1] + (pt->grav[1] * grav);
		pt->accel[2] = pt->dvel[2] + (pt->grav[2] * grav);
	}

	// move dead emitters at the start of the list back to free
	for (;;)
	{
		emitter_t *kill = active_emitters;

		if (kill && !kill->particles)
		{
			active_emitters = kill->next;
			kill->next = free_emitters;
			free_emitters = kill;
			continue;
		}

		break;
	}

	for (e = active_emitters; e; e = e->next)
	{
		// move dead emitters in the middle of the list back to free
		for (;;)
		{
			emitter_t *kill = e->next;

			if (kill && !kill->particles)
			{
				e->next = kill->next;
				kill->next = free_emitters;
				free_emitters = kill;
				continue;
			}

			break;
		}

		// run this emitter
		CL_RunEmitter (e);

		if (e->particles)
		{
			// add this emitter to the draw lists
			// R_AddAlphaParticles (e);
		}
	}
}



